@@ -30,7 +30,6 @@ module Network.TLS.Packet
     , decodeHandshake
     , decodeDeprecatedHandshake
     , encodeHandshake
-    , encodeHandshakes
     , encodeHandshakeHeader
     , encodeHandshakeContent
 
@@ -47,33 +46,40 @@ module Network.TLS.Packet
 
     -- * generate things for packet content
     , generateMasterSecret
+    , generateExtendedMasterSec
     , generateKeyBlock
     , generateClientFinished
     , generateServerFinished
 
     , generateCertificateVerify_SSL
+    , generateCertificateVerify_SSL_DSS
 
     -- * for extensions parsing
     , getSignatureHashAlgorithm
     , putSignatureHashAlgorithm
+    , getBinaryVersion
+    , putBinaryVersion
+    , getClientRandom32
+    , putClientRandom32
+    , getServerRandom32
+    , putServerRandom32
+    , getExtensions
+    , putExtension
+    , getSession
+    , putSession
+    , putDNames
+    , getDNames
     ) where
 
+import Network.TLS.Imports
 import Network.TLS.Struct
 import Network.TLS.Wire
 import Network.TLS.Cap
-import Data.Maybe (fromJust)
-import Data.Word
-import Control.Applicative ((<$>), (<*>))
-import Control.Monad
-import Data.ASN1.Types (fromASN1, toASN1)
-import Data.ASN1.Encoding (decodeASN1', encodeASN1')
-import Data.ASN1.BinaryEncoding (DER(..))
 import Data.X509 (CertificateChainRaw(..), encodeCertificateChain, decodeCertificateChain)
 import Network.TLS.Crypto
 import Network.TLS.MAC
-import Network.TLS.Cipher (CipherKeyExchangeType(..))
-import Network.TLS.Util.Serialization (os2ip,i2ospOf_)
-import Data.ByteString (ByteString)
+import Network.TLS.Cipher (CipherKeyExchangeType(..), Cipher(..))
+import Network.TLS.Util.ASN1
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as BC
 import           Data.ByteArray (ByteArrayAccess)
@@ -82,7 +88,6 @@ import qualified Data.ByteArray as B (convert)
 data CurrentParams = CurrentParams
     { cParamsVersion     :: Version                     -- ^ current protocol version
     , cParamsKeyXchgType :: Maybe CipherKeyExchangeType -- ^ current key exchange type
-    , cParamsSupportNPN  :: Bool                        -- ^ support Next Protocol Negotiation extension
     } deriving (Show,Eq)
 
 {- marshall helpers -}
@@ -94,8 +99,14 @@ getVersion = do
         Nothing -> fail ("invalid version : " ++ show major ++ "," ++ show minor)
         Just v  -> return v
 
-putVersion :: Version -> Put
-putVersion ver = putWord8 major >> putWord8 minor
+getBinaryVersion :: Get (Maybe Version)
+getBinaryVersion = do
+    major <- getWord8
+    minor <- getWord8
+    return $ verOfNum (major, minor)
+
+putBinaryVersion :: Version -> Put
+putBinaryVersion ver = putWord8 major >> putWord8 minor
   where (major, minor) = numericalVer ver
 
 getHeaderType :: Get ProtocolType
@@ -119,7 +130,7 @@ getHandshakeType = do
  - decode and encode headers
  -}
 decodeHeader :: ByteString -> Either TLSError Header
-decodeHeader = runGetErr "header" $ liftM3 Header getHeaderType getVersion getWord16
+decodeHeader = runGetErr "header" $ Header <$> getHeaderType <*> getVersion <*> getWord16
 
 decodeDeprecatedHeaderLength :: ByteString -> Either TLSError Word16
 decodeDeprecatedHeaderLength = runGetErr "deprecatedheaderlength" $ subtract 0x8000 <$> getWord16
@@ -132,7 +143,7 @@ decodeDeprecatedHeader size =
         return $ Header ProtocolType_DeprecatedHandshake version size
 
 encodeHeader :: Header -> ByteString
-encodeHeader (Header pt ver len) = runPut (putHeaderType pt >> putVersion ver >> putWord16 len)
+encodeHeader (Header pt ver len) = runPut (putHeaderType pt >> putBinaryVersion ver >> putWord16 len)
         {- FIXME check len <= 2^14 -}
 
 encodeHeaderNoVer :: Header -> ByteString
@@ -152,19 +163,19 @@ decodeAlert = do
         (_, Nothing)     -> fail "cannot decode alert description"
 
 decodeAlerts :: ByteString -> Either TLSError [(AlertLevel, AlertDescription)]
-decodeAlerts = runGetErr "alerts" $ loop
+decodeAlerts = runGetErr "alerts" loop
   where loop = do
             r <- remaining
             if r == 0
                 then return []
-                else liftM2 (:) decodeAlert loop
+                else (:) <$> decodeAlert <*> loop
 
 encodeAlerts :: [(AlertLevel, AlertDescription)] -> ByteString
 encodeAlerts l = runPut $ mapM_ encodeAlert l
   where encodeAlert (al, ad) = putWord8 (valOfType al) >> putWord8 (valOfType ad)
 
 {- decode and encode HANDSHAKE -}
-decodeHandshakeRecord :: ByteString -> GetResult (HandshakeType, Bytes)
+decodeHandshakeRecord :: ByteString -> GetResult (HandshakeType, ByteString)
 decodeHandshakeRecord = runGet "handshake-record" $ do
     ty      <- getHandshakeType
     content <- getOpaque24
@@ -182,9 +193,6 @@ decodeHandshake cp ty = runGetErr ("handshake[" ++ show ty ++ "]") $ case ty of
     HandshakeType_CertVerify      -> decodeCertVerify cp
     HandshakeType_ClientKeyXchg   -> decodeClientKeyXchg cp
     HandshakeType_Finished        -> decodeFinished
-    HandshakeType_NPN             -> do
-        unless (cParamsSupportNPN cp) $ fail "unsupported handshake type"
-        decodeNextProtocolNegotiation
 
 decodeDeprecatedHandshake :: ByteString -> Either TLSError Handshake
 decodeDeprecatedHandshake b = runGetErr "deprecatedhandshake" getDeprecated b
@@ -220,7 +228,7 @@ decodeClientHello = do
     compressions <- getWords8
     r            <- remaining
     exts <- if hasHelloExtensions ver && r > 0
-            then fmap fromIntegral getWord16 >>= getExtensions
+            then fromIntegral <$> getWord16 >>= getExtensions
             else return []
     return $ ClientHello ver random session ciphers compressions exts Nothing
 
@@ -233,7 +241,7 @@ decodeServerHello = do
     compressionid <- getWord8
     r             <- remaining
     exts <- if hasHelloExtensions ver && r > 0
-            then fmap fromIntegral getWord16 >>= getExtensions
+            then fromIntegral <$> getWord16 >>= getExtensions
             else return []
     return $ ServerHello ver random session cipherid compressionid exts
 
@@ -251,34 +259,29 @@ decodeCertificates = do
 decodeFinished :: Get Handshake
 decodeFinished = Finished <$> (remaining >>= getBytes)
 
-decodeNextProtocolNegotiation :: Get Handshake
-decodeNextProtocolNegotiation = do
-    opaque <- getOpaque8
-    _      <- getOpaque8 -- ignore padding
-    return $ HsNextProtocolNegotiation opaque
-
 decodeCertRequest :: CurrentParams -> Get Handshake
 decodeCertRequest cp = do
-    certTypes <- map (fromJust . valToType . fromIntegral) <$> getWords8
-
+    mcertTypes <- map (valToType . fromIntegral) <$> getWords8
+    certTypes <- mapM (fromJustM "decodeCertRequest") mcertTypes
     sigHashAlgs <- if cParamsVersion cp >= TLS12
                        then Just <$> (getWord16 >>= getSignatureHashAlgorithms)
                        else return Nothing
+    CertRequest certTypes sigHashAlgs <$> getDNames
+  where getSignatureHashAlgorithms len = getList (fromIntegral len) (getSignatureHashAlgorithm >>= \sh -> return (2, sh))
+
+-- | Decode a list CA distinguished names
+getDNames :: Get [DistinguishedName]
+getDNames = do
     dNameLen <- getWord16
     -- FIXME: Decide whether to remove this check completely or to make it an option.
     -- when (cParamsVersion cp < TLS12 && dNameLen < 3) $ fail "certrequest distinguishname not of the correct size"
-    dNames <- getList (fromIntegral dNameLen) getDName
-    return $ CertRequest certTypes sigHashAlgs dNames
-  where getSignatureHashAlgorithms len = getList (fromIntegral len) (getSignatureHashAlgorithm >>= \sh -> return (2, sh))
-        getDName = do
-            dName <- getOpaque16
-            when (B.length dName == 0) $ fail "certrequest: invalid DN length"
-            dn <- case decodeASN1' DER dName of
-                    Left e      -> fail ("cert request decoding DistinguishedName ASN1 failed: " ++ show e)
-                    Right asn1s -> case fromASN1 asn1s of
-                                        Left e      -> fail ("cert request parsing DistinguishedName ASN1 failed: " ++ show e)
-                                        Right (d,_) -> return d
-            return (2 + B.length dName, dn)
+    getList (fromIntegral dNameLen) getDName
+  where
+    getDName = do
+        dName <- getOpaque16
+        when (B.length dName == 0) $ fail "certrequest: invalid DN length"
+        dn <- either fail return $ decodeASN1Object "cert request DistinguishedName" dName
+        return (2 + B.length dName, dn)
 
 decodeCertVerify :: CurrentParams -> Get Handshake
 decodeCertVerify cp = CertVerify <$> getDigitallySigned (cParamsVersion cp)
@@ -292,22 +295,11 @@ decodeClientKeyXchg cp = -- case  ClientKeyXchg <$> (remaining >>= getBytes)
         parseCKE CipherKeyExchange_DHE_RSA = parseClientDHPublic
         parseCKE CipherKeyExchange_DHE_DSS = parseClientDHPublic
         parseCKE CipherKeyExchange_DH_Anon = parseClientDHPublic
-        parseCKE CipherKeyExchange_ECDHE_RSA = parseClientECDHPublic
+        parseCKE CipherKeyExchange_ECDHE_RSA   = parseClientECDHPublic
         parseCKE CipherKeyExchange_ECDHE_ECDSA = parseClientECDHPublic
         parseCKE _                         = error "unsupported client key exchange type"
         parseClientDHPublic = CKX_DH . dhPublic <$> getInteger16
-        parseClientECDHPublic = do
-            len      <- getWord8
-            formatTy <- getWord8
-            case formatTy of
-                4 -> do -- uncompressed
-                    let siz = fromIntegral len `div` 2
-                    xb <- getBytes siz
-                    yb <- getBytes siz
-                    let x = os2ip xb
-                        y = os2ip yb
-                    return $ CKX_ECDH $ ecdhPublic x y siz
-                _ -> error ("unsupported EC format type: " ++ show formatTy)
+        parseClientECDHPublic = CKX_ECDH <$> getOpaque8
 
 decodeServerKeyXchg_DH :: Get ServerDHParams
 decodeServerKeyXchg_DH = getServerDHParams
@@ -335,13 +327,13 @@ decodeServerKeyXchgAlgorithmData ver cke = toCKE
                 signature <- getDigitallySigned ver
                 return $ SKX_DHE_DSS dhparams signature
             CipherKeyExchange_ECDHE_RSA -> do
-                dhparams  <- getServerECDHParams
+                ecdhparams  <- getServerECDHParams
                 signature <- getDigitallySigned ver
-                return $ SKX_ECDHE_RSA dhparams signature
+                return $ SKX_ECDHE_RSA ecdhparams signature
             CipherKeyExchange_ECDHE_ECDSA -> do
-                dhparams  <- getServerECDHParams
+                ecdhparams  <- getServerECDHParams
                 signature <- getDigitallySigned ver
-                return $ SKX_ECDHE_ECDSA dhparams signature
+                return $ SKX_ECDHE_ECDSA ecdhparams signature
             _ -> do
                 bs <- remaining >>= getBytes
                 return $ SKX_Unknown bs
@@ -355,15 +347,12 @@ decodeServerKeyXchg cp =
 encodeHandshake :: Handshake -> ByteString
 encodeHandshake o =
     let content = runPut $ encodeHandshakeContent o in
-    let len = fromIntegral $ B.length content in
+    let len = B.length content in
     let header = case o of
                     ClientHello _ _ _ _ _ _ (Just _) -> "" -- SSLv2 ClientHello message
                     _ -> runPut $ encodeHandshakeHeader (typeOfHandshake o) len in
     B.concat [ header, content ]
 
-encodeHandshakes :: [Handshake] -> ByteString
-encodeHandshakes hss = B.concat $ map encodeHandshake hss
-
 encodeHandshakeHeader :: HandshakeType -> Int -> Put
 encodeHandshakeHeader ty len = putWord8 (valOfType ty) >> putWord24 len
 
@@ -372,7 +361,7 @@ encodeHandshakeContent :: Handshake -> Put
 encodeHandshakeContent (ClientHello _ _ _ _ _ _ (Just deprecated)) = do
     putBytes deprecated
 encodeHandshakeContent (ClientHello version random session cipherIDs compressionIDs exts Nothing) = do
-    putVersion version
+    putBinaryVersion version
     putClientRandom32 random
     putSession session
     putWords16 cipherIDs
@@ -380,10 +369,14 @@ encodeHandshakeContent (ClientHello version random session cipherIDs compression
     putExtensions exts
     return ()
 
-encodeHandshakeContent (ServerHello version random session cipherID compressionID exts) =
-    putVersion version >> putServerRandom32 random >> putSession session
-                       >> putWord16 cipherID >> putWord8 compressionID
-                       >> putExtensions exts >> return ()
+encodeHandshakeContent (ServerHello version random session cipherid compressionID exts) = do
+    putBinaryVersion version
+    putServerRandom32 random
+    putSession session
+    putWord16 cipherid
+    putWord8 compressionID
+    putExtensions exts
+    return ()
 
 encodeHandshakeContent (Certificates cc) = putOpaque24 (runPut $ mapM_ putOpaque24 certs)
   where (CertificateChainRaw certs) = encodeCertificateChain cc
@@ -392,11 +385,7 @@ encodeHandshakeContent (ClientKeyXchg ckx) = do
     case ckx of
         CKX_RSA encryptedPreMaster -> putBytes encryptedPreMaster
         CKX_DH clientDHPublic      -> putInteger16 $ dhUnwrapPublic clientDHPublic
-        CKX_ECDH clientECDHPublic  -> do
-            let (x,y,siz) = ecdhUnwrapPublic clientECDHPublic
-            let xb = i2ospOf_ siz x
-                yb = i2ospOf_ siz y
-            putOpaque8 $ B.concat [B.singleton 4,xb,yb]
+        CKX_ECDH bytes             -> putOpaque8 bytes
 
 encodeHandshakeContent (ServerKeyXchg skg) =
     case skg of
@@ -409,36 +398,35 @@ encodeHandshakeContent (ServerKeyXchg skg) =
         SKX_Unparsed bytes     -> putBytes bytes
         _                      -> error ("encodeHandshakeContent: cannot handle: " ++ show skg)
 
-encodeHandshakeContent (HelloRequest) = return ()
-encodeHandshakeContent (ServerHelloDone) = return ()
+encodeHandshakeContent HelloRequest    = return ()
+encodeHandshakeContent ServerHelloDone = return ()
 
 encodeHandshakeContent (CertRequest certTypes sigAlgs certAuthorities) = do
     putWords8 (map valOfType certTypes)
     case sigAlgs of
         Nothing -> return ()
-        Just l  -> putWords16 $ map (\(x,y) -> (fromIntegral $ valOfType x) * 256 + (fromIntegral $ valOfType y)) l
-    encodeCertAuthorities certAuthorities
-  where -- Convert a distinguished name to its DER encoding.
-        encodeCA dn = return $ encodeASN1' DER (toASN1 dn []) --B.concat $ L.toChunks $ encodeDN dn
-
-        -- Encode a list of distinguished names.
-        encodeCertAuthorities certAuths = do
-            enc <- mapM encodeCA certAuths
-            let totLength = sum $ map (((+) 2) . B.length) enc
-            putWord16 (fromIntegral totLength)
-            mapM_ (\ b -> putWord16 (fromIntegral (B.length b)) >> putBytes b) enc
+        Just l  -> putWords16 $ map (\(x,y) -> fromIntegral (valOfType x) * 256 + fromIntegral (valOfType y)) l
+    putDNames certAuthorities
 
 encodeHandshakeContent (CertVerify digitallySigned) = putDigitallySigned digitallySigned
 
 encodeHandshakeContent (Finished opaque) = putBytes opaque
 
-encodeHandshakeContent (HsNextProtocolNegotiation protocol) = do
-    putOpaque8 protocol
-    putOpaque8 $ B.replicate paddingLen 0
-  where paddingLen = 32 - ((B.length protocol + 2) `mod` 32)
+------------------------------------------------------------
+
+-- | Encode a list of distinguished names.
+putDNames :: [DistinguishedName] -> Put
+putDNames dnames = do
+    enc <- mapM encodeCA dnames
+    let totLength = sum $ map ((+) 2 . B.length) enc
+    putWord16 (fromIntegral totLength)
+    mapM_ (\ b -> putWord16 (fromIntegral (B.length b)) >> putBytes b) enc
+  where
+    -- Convert a distinguished name to its DER encoding.
+    encodeCA dn = return $ encodeASN1Object dn
 
 {- FIXME make sure it return error if not 32 available -}
-getRandom32 :: Get Bytes
+getRandom32 :: Get ByteString
 getRandom32 = getBytes 32
 
 getServerRandom32 :: Get ServerRandom
@@ -447,7 +435,7 @@ getServerRandom32 = ServerRandom <$> getRandom32
 getClientRandom32 :: Get ClientRandom
 getClientRandom32 = ClientRandom <$> getRandom32
 
-putRandom32 :: Bytes -> Put
+putRandom32 :: ByteString -> Put
 putRandom32 = putBytes
 
 putClientRandom32 :: ClientRandom -> Put
@@ -474,10 +462,10 @@ getExtensions len = do
     extdatalen <- getWord16
     extdata <- getBytes $ fromIntegral extdatalen
     extxs <- getExtensions (len - fromIntegral extdatalen - 4)
-    return $ (extty, extdata) : extxs
+    return $ ExtensionRaw extty extdata : extxs
 
 putExtension :: ExtensionRaw -> Put
-putExtension (ty, l) = putWord16 ty >> putOpaque16 l
+putExtension (ExtensionRaw ty l) = putWord16 ty >> putOpaque16 l
 
 putExtensions :: [ExtensionRaw] -> Put
 putExtensions [] = return ()
@@ -485,8 +473,8 @@ putExtensions es = putOpaque16 (runPut $ mapM_ putExtension es)
 
 getSignatureHashAlgorithm :: Get HashAndSignatureAlgorithm
 getSignatureHashAlgorithm = do
-    h <- fromJust . valToType <$> getWord8
-    s <- fromJust . valToType <$> getWord8
+    h <- (valToType <$> getWord8) >>= fromJustM "getSignatureHashAlgorithm"
+    s <- (valToType <$> getWord8) >>= fromJustM "getSignatureHashAlgorithm"
     return (h,s)
 
 putSignatureHashAlgorithm :: HashAndSignatureAlgorithm -> Put
@@ -499,40 +487,29 @@ getServerDHParams = ServerDHParams <$> getBigNum16 <*> getBigNum16 <*> getBigNum
 putServerDHParams :: ServerDHParams -> Put
 putServerDHParams (ServerDHParams p g y) = mapM_ putBigNum16 [p,g,y]
 
+-- RFC 4492 Section 5.4 Server Key Exchange
 getServerECDHParams :: Get ServerECDHParams
 getServerECDHParams = do
     curveType <- getWord8
     case curveType of
-        1 -> do -- explicit prime
-            _prime    <- getOpaque8
-            _a        <- getOpaque8
-            _b        <- getOpaque8
-            _base     <- getOpaque8
-            _order    <- getOpaque8
-            _cofactor <- getOpaque8
-            error "cannot handle explicit prime ECDH Params"
-        2 -> -- explicit_char2
-            error "cannot handle explicit char2 ECDH Params"
-        3 -> do -- ECParameters ECCurveType: curve name type
-            w16 <- getWord16   -- ECParameters NamedCurve
-            mxy <- getOpaque8 -- ECPoint
-            let xy = B.drop 1 mxy
-                siz = B.length xy `div` 2
-                (xb,yb) = B.splitAt siz xy
-                x = os2ip xb
-                y = os2ip yb
-            return $ ServerECDHParams (ecdhParams w16) (ecdhPublic x y siz)
+        3 -> do               -- ECParameters ECCurveType: curve name type
+            mgrp <- toEnumSafe16 <$> getWord16  -- ECParameters NamedCurve
+            case mgrp of
+              Nothing -> error "getServerECDHParams: unknown group"
+              Just grp -> do
+                  mxy <- getOpaque8 -- ECPoint
+                  case decodeGroupPublic grp mxy of
+                    Left e       -> error $ "getServerECDHParams: " ++ show e
+                    Right grppub -> return $ ServerECDHParams grp grppub
         _ ->
-            error "unknown type for ECDH Params"
+            error "getServerECDHParams: unknown type for ECDH Params"
 
+-- RFC 4492 Section 5.4 Server Key Exchange
 putServerECDHParams :: ServerECDHParams -> Put
-putServerECDHParams (ServerECDHParams ecdhparams ecdhpub) = do
-    let (w16,x,y,siz) = ecdhUnwrap ecdhparams ecdhpub
-    putWord8 3    -- ECParameters ECCurveType: curve name type
-    putWord16 w16 -- ECParameters NamedCurve
-    let xb = i2ospOf_ siz x
-        yb = i2ospOf_ siz y
-    putOpaque8 $ B.concat [B.singleton 4,xb,yb] -- ECPoint
+putServerECDHParams (ServerECDHParams grp grppub) = do
+    putWord8 3                            -- ECParameters ECCurveType
+    putWord16 $ fromEnumSafe16 grp        -- ECParameters NamedCurve
+    putOpaque8 $ encodeGroupPublic grppub -- ECPoint
 
 getDigitallySigned :: Version -> Get DigitallySigned
 getDigitallySigned ver
@@ -557,12 +534,12 @@ encodeChangeCipherSpec :: ByteString
 encodeChangeCipherSpec = runPut (putWord8 1)
 
 -- rsa pre master secret
-decodePreMasterSecret :: Bytes -> Either TLSError (Version, Bytes)
-decodePreMasterSecret = runGetErr "pre-master-secret" $ do
-    liftM2 (,) getVersion (getBytes 46)
+decodePreMasterSecret :: ByteString -> Either TLSError (Version, ByteString)
+decodePreMasterSecret = runGetErr "pre-master-secret" $
+    (,) <$> getVersion <*> getBytes 46
 
-encodePreMasterSecret :: Version -> Bytes -> Bytes
-encodePreMasterSecret version bytes = runPut (putVersion version >> putBytes bytes)
+encodePreMasterSecret :: Version -> ByteString -> ByteString
+encodePreMasterSecret version bytes = runPut (putBinaryVersion version >> putBytes bytes)
 
 -- | in certain cases, we haven't manage to decode ServerKeyExchange properly,
 -- because the decoding was too eager and the cipher wasn't been set yet.
@@ -570,7 +547,7 @@ encodePreMasterSecret version bytes = runPut (putVersion version >> putBytes byt
 -- able to really decode the server key xchange if it's unparsed.
 decodeReallyServerKeyXchgAlgorithmData :: Version
                                        -> CipherKeyExchangeType
-                                       -> Bytes
+                                       -> ByteString
                                        -> Either TLSError ServerKeyXchgAlgorithmData
 decodeReallyServerKeyXchgAlgorithmData ver cke =
     runGetErr "server-key-xchg-algorithm-data" (decodeServerKeyXchgAlgorithmData ver cke)
@@ -579,49 +556,75 @@ decodeReallyServerKeyXchgAlgorithmData ver cke =
 {-
  - generate things for packet content
  -}
-type PRF = Bytes -> Bytes -> Int -> Bytes
+type PRF = ByteString -> ByteString -> Int -> ByteString
 
-generateMasterSecret_SSL :: ByteArrayAccess preMaster => preMaster -> ClientRandom -> ServerRandom -> Bytes
+-- | The TLS12 PRF is cipher specific, and some TLS12 algorithms use SHA384
+-- instead of the default SHA256.
+getPRF :: Version -> Cipher -> PRF
+getPRF ver ciph
+    | ver < TLS12 = prf_MD5SHA1
+    | maybe True (< TLS12) (cipherMinVer ciph) = prf_SHA256
+    | otherwise = prf_TLS ver $ fromMaybe SHA256 $ cipherPRFHash ciph
+
+generateMasterSecret_SSL :: ByteArrayAccess preMaster => preMaster -> ClientRandom -> ServerRandom -> ByteString
 generateMasterSecret_SSL premasterSecret (ClientRandom c) (ServerRandom s) =
-    B.concat $ map (computeMD5) ["A","BB","CCC"]
+    B.concat $ map computeMD5 ["A","BB","CCC"]
   where computeMD5  label = hash MD5 $ B.concat [ B.convert premasterSecret, computeSHA1 label ]
         computeSHA1 label = hash SHA1 $ B.concat [ label, B.convert premasterSecret, c, s ]
 
-generateMasterSecret_TLS :: ByteArrayAccess preMaster => PRF -> preMaster -> ClientRandom -> ServerRandom -> Bytes
+generateMasterSecret_TLS :: ByteArrayAccess preMaster => PRF -> preMaster -> ClientRandom -> ServerRandom -> ByteString
 generateMasterSecret_TLS prf premasterSecret (ClientRandom c) (ServerRandom s) =
     prf (B.convert premasterSecret) seed 48
   where seed = B.concat [ "master secret", c, s ]
 
-generateMasterSecret :: ByteArrayAccess preMaster => Version -> preMaster -> ClientRandom -> ServerRandom -> Bytes
-generateMasterSecret SSL2  = generateMasterSecret_SSL
-generateMasterSecret SSL3  = generateMasterSecret_SSL
-generateMasterSecret TLS10 = generateMasterSecret_TLS prf_MD5SHA1
-generateMasterSecret TLS11 = generateMasterSecret_TLS prf_MD5SHA1
-generateMasterSecret TLS12 = generateMasterSecret_TLS prf_SHA256
+generateMasterSecret :: ByteArrayAccess preMaster
+                     => Version
+                     -> Cipher
+                     -> preMaster
+                     -> ClientRandom
+                     -> ServerRandom
+                     -> ByteString
+generateMasterSecret SSL2 _ = generateMasterSecret_SSL
+generateMasterSecret SSL3 _ = generateMasterSecret_SSL
+generateMasterSecret v    c = generateMasterSecret_TLS $ getPRF v c
 
-generateKeyBlock_TLS :: PRF -> ClientRandom -> ServerRandom -> Bytes -> Int -> Bytes
+generateExtendedMasterSec :: ByteArrayAccess preMaster
+                          => Version
+                          -> Cipher
+                          -> preMaster
+                          -> ByteString
+                          -> ByteString
+generateExtendedMasterSec v c premasterSecret sessionHash =
+    getPRF v c (B.convert premasterSecret) seed 48
+  where seed = B.append "extended master secret" sessionHash
+
+generateKeyBlock_TLS :: PRF -> ClientRandom -> ServerRandom -> ByteString -> Int -> ByteString
 generateKeyBlock_TLS prf (ClientRandom c) (ServerRandom s) mastersecret kbsize =
     prf mastersecret seed kbsize where seed = B.concat [ "key expansion", s, c ]
 
-generateKeyBlock_SSL :: ClientRandom -> ServerRandom -> Bytes -> Int -> Bytes
+generateKeyBlock_SSL :: ClientRandom -> ServerRandom -> ByteString -> Int -> ByteString
 generateKeyBlock_SSL (ClientRandom c) (ServerRandom s) mastersecret kbsize =
     B.concat $ map computeMD5 $ take ((kbsize `div` 16) + 1) labels
   where labels            = [ uncurry BC.replicate x | x <- zip [1..] ['A'..'Z'] ]
         computeMD5  label = hash MD5 $ B.concat [ mastersecret, computeSHA1 label ]
         computeSHA1 label = hash SHA1 $ B.concat [ label, mastersecret, s, c ]
 
-generateKeyBlock :: Version -> ClientRandom -> ServerRandom -> Bytes -> Int -> Bytes
-generateKeyBlock SSL2  = generateKeyBlock_SSL
-generateKeyBlock SSL3  = generateKeyBlock_SSL
-generateKeyBlock TLS10 = generateKeyBlock_TLS prf_MD5SHA1
-generateKeyBlock TLS11 = generateKeyBlock_TLS prf_MD5SHA1
-generateKeyBlock TLS12 = generateKeyBlock_TLS prf_SHA256
+generateKeyBlock :: Version
+                 -> Cipher
+                 -> ClientRandom
+                 -> ServerRandom
+                 -> ByteString
+                 -> Int
+                 -> ByteString
+generateKeyBlock SSL2 _ = generateKeyBlock_SSL
+generateKeyBlock SSL3 _ = generateKeyBlock_SSL
+generateKeyBlock v    c = generateKeyBlock_TLS $ getPRF v c
 
-generateFinished_TLS :: PRF -> Bytes -> Bytes -> HashCtx -> Bytes
+generateFinished_TLS :: PRF -> ByteString -> ByteString -> HashCtx -> ByteString
 generateFinished_TLS prf label mastersecret hashctx = prf mastersecret seed 12
   where seed = B.concat [ label, hashFinal hashctx ]
 
-generateFinished_SSL :: Bytes -> Bytes -> HashCtx -> Bytes
+generateFinished_SSL :: ByteString -> ByteString -> HashCtx -> ByteString
 generateFinished_SSL sender mastersecret hashctx = B.concat [md5hash, sha1hash]
   where md5hash  = hash MD5 $ B.concat [ mastersecret, pad2, md5left ]
         sha1hash = hash SHA1 $ B.concat [ mastersecret, B.take 40 pad2, sha1left ]
@@ -632,28 +635,49 @@ generateFinished_SSL sender mastersecret hashctx = B.concat [md5hash, sha1hash]
         pad2     = B.replicate 48 0x5c
         pad1     = B.replicate 48 0x36
 
-generateClientFinished :: Version -> Bytes -> HashCtx -> Bytes
-generateClientFinished ver
+generateClientFinished :: Version
+                       -> Cipher
+                       -> ByteString
+                       -> HashCtx
+                       -> ByteString
+generateClientFinished ver ciph
     | ver < TLS10 = generateFinished_SSL "CLNT"
-    | ver < TLS12 = generateFinished_TLS prf_MD5SHA1 "client finished"
-    | otherwise   = generateFinished_TLS prf_SHA256 "client finished"
+    | otherwise   = generateFinished_TLS (getPRF ver ciph) "client finished"
 
-generateServerFinished :: Version -> Bytes -> HashCtx -> Bytes
-generateServerFinished ver
+generateServerFinished :: Version
+                       -> Cipher
+                       -> ByteString
+                       -> HashCtx
+                       -> ByteString
+generateServerFinished ver ciph
     | ver < TLS10 = generateFinished_SSL "SRVR"
-    | ver < TLS12 = generateFinished_TLS prf_MD5SHA1 "server finished"
-    | otherwise   = generateFinished_TLS prf_SHA256 "server finished"
+    | otherwise   = generateFinished_TLS (getPRF ver ciph) "server finished"
 
-generateCertificateVerify_SSL :: Bytes -> HashCtx -> Bytes
+{- returns *output* after final MD5/SHA1 -}
+generateCertificateVerify_SSL :: ByteString -> HashCtx -> ByteString
 generateCertificateVerify_SSL = generateFinished_SSL ""
 
-encodeSignedDHParams :: ServerDHParams -> ClientRandom -> ServerRandom -> Bytes
+{- returns *input* before final SHA1 -}
+generateCertificateVerify_SSL_DSS :: ByteString -> HashCtx -> ByteString
+generateCertificateVerify_SSL_DSS mastersecret hashctx = toHash
+  where toHash = B.concat [ mastersecret, pad2, sha1left ]
+
+        sha1left = hashFinal $ flip hashUpdate pad1
+                             $ hashUpdate hashctx mastersecret
+        pad2     = B.replicate 40 0x5c
+        pad1     = B.replicate 40 0x36
+
+encodeSignedDHParams :: ServerDHParams -> ClientRandom -> ServerRandom -> ByteString
 encodeSignedDHParams dhparams cran sran = runPut $
     putClientRandom32 cran >> putServerRandom32 sran >> putServerDHParams dhparams
 
 -- Combination of RFC 5246 and 4492 is ambiguous.
 -- Let's assume ecdhe_rsa and ecdhe_dss are identical to
 -- dhe_rsa and dhe_dss.
-encodeSignedECDHParams :: ServerECDHParams -> ClientRandom -> ServerRandom -> Bytes
+encodeSignedECDHParams :: ServerECDHParams -> ClientRandom -> ServerRandom -> ByteString
 encodeSignedECDHParams dhparams cran sran = runPut $
     putClientRandom32 cran >> putServerRandom32 sran >> putServerECDHParams dhparams
+
+fromJustM :: MonadFail m => String -> Maybe a -> m a
+fromJustM what Nothing  = fail ("fromJustM " ++ what ++ ": Nothing")
+fromJustM _    (Just x) = return x
